@@ -37,6 +37,8 @@
 #include <stdarg.h>
 #include <math.h>
 
+#include "config.h"
+
 #if RANDR_MAJOR > 1 || (RANDR_MAJOR == 1 && RANDR_MINOR >= 2)
 #define HAS_RANDR_1_2 1
 #endif
@@ -48,6 +50,8 @@ static int	screen = -1;
 static Bool	verbose = False;
 static Bool	automatic = False;
 static Bool	properties = False;
+static Bool	grab_server = True;
+static Bool	no_primary = False;
 
 static char *direction[5] = {
     "normal", 
@@ -106,6 +110,7 @@ usage(void)
     fprintf(stderr, "  --screen <screen>\n");
     fprintf(stderr, "  --verbose\n");
     fprintf(stderr, "  --dryrun\n");
+    fprintf(stderr, "  --nograb\n");
 #if HAS_RANDR_1_2
     fprintf(stderr, "  --prop or --properties\n");
     fprintf(stderr, "  --fb <width>x<height>\n");
@@ -134,6 +139,9 @@ usage(void)
     fprintf(stderr, "      --off\n");
     fprintf(stderr, "      --crtc <crtc>\n");
     fprintf(stderr, "      --panning <w>x<h>[+<x>+<y>[/<track:w>x<h>+<x>+<y>[/<border:l>/<t>/<r>/<b>]]]\n");
+    fprintf(stderr, "      --gamma <r>:<g>:<b>\n");
+    fprintf(stderr, "      --primary\n");
+    fprintf(stderr, "  --noprimary\n");
     fprintf(stderr, "  --newmode <name> <clock MHz>\n");
     fprintf(stderr, "            <hdisp> <hsync-start> <hsync-end> <htotal>\n");
     fprintf(stderr, "            <vdisp> <vsync-start> <vsync-end> <vtotal>\n");
@@ -235,6 +243,8 @@ typedef enum _changes {
     changes_property = (1 << 8),
     changes_transform = (1 << 9),
     changes_panning = (1 << 10),
+    changes_gamma = (1 << 11),
+    changes_primary = (1 << 12),
 } changes_t;
 
 typedef enum _name_kind {
@@ -311,11 +321,19 @@ struct _output {
 
     int		    x, y;
     Rotation	    rotation;
-    
+
     XRRPanning      panning;
 
     Bool    	    automatic;
     transform_t	    transform;
+
+    struct {
+	float red;
+	float green;
+	float blue;
+    } gamma;
+
+    Bool	    primary;
 };
 
 typedef enum _umode_action {
@@ -395,13 +413,13 @@ mode_width (XRRModeInfo *mode_info, Rotation rotation)
     }
 }
 
-Bool
+static Bool
 transform_point (XTransform *transform, double *xp, double *yp)
 {
     double  vector[3];
     double  result[3];
     int	    i, j;
-    double  partial, v;
+    double  v;
 
     vector[0] = *xp;
     vector[1] = *yp;
@@ -424,7 +442,7 @@ transform_point (XTransform *transform, double *xp, double *yp)
     return True;
 }
 
-Bool
+static void
 path_bounds (XTransform *transform, point_t *points, int npoints, box_t *box)
 {
     int	    i;
@@ -612,7 +630,7 @@ add_output (void)
     output_t *output = calloc (1, sizeof (output_t));
 
     if (!output)
-	fatal ("out of memory");
+	fatal ("out of memory\n");
     output->next = NULL;
     *outputs_tail = output;
     outputs_tail = &output->next;
@@ -736,6 +754,7 @@ find_mode_by_xid (RRMode mode)
     return find_mode (&mode_name, 0);
 }
 
+#if 0
 static XRRModeInfo *
 find_mode_by_name (char *name)
 {
@@ -744,6 +763,7 @@ find_mode_by_name (char *name)
     set_name_string (&mode_name, name);
     return find_mode (&mode_name, 0);
 }
+#endif
 
 static
 XRRModeInfo *
@@ -757,7 +777,7 @@ find_mode_for_output (output_t *output, name_t *name)
     for (m = 0; m < output_info->nmode; m++)
     {
 	XRRModeInfo	    *mode;
-	
+
 	mode = find_mode_by_xid (output_info->modes[m]);
 	if (!mode) continue;
 	if ((name->kind & name_xid) && name->xid == mode->id)
@@ -768,7 +788,11 @@ find_mode_for_output (output_t *output, name_t *name)
 	if ((name->kind & name_string) && !strcmp (name->string, mode->name))
 	{
 	    float   dist;
-	    
+
+	    /* Stay away from doublescan modes unless refresh rate is specified. */
+	    if (!output->refresh && (mode->modeFlags & RR_DoubleScan))
+		continue;
+
 	    if (output->refresh)
 		dist = fabs (mode_refresh (mode) - output->refresh);
 	    else
@@ -783,7 +807,7 @@ find_mode_for_output (output_t *output, name_t *name)
     return best;
 }
 
-XRRModeInfo *
+static XRRModeInfo *
 preferred_mode (output_t *output)
 {
     XRROutputInfo   *output_info = output->output_info;
@@ -851,6 +875,7 @@ crtc_can_use_rotation (crtc_t *crtc, Rotation rotation)
     return False;
 }
 
+#if 0
 static Bool
 crtc_can_use_transform (crtc_t *crtc, XTransform *transform)
 {
@@ -861,6 +886,7 @@ crtc_can_use_transform (crtc_t *crtc, XTransform *transform)
 	return True;
     return False;
 }
+#endif
 
 /*
  * Report only rotations that are supported by all crtcs
@@ -908,6 +934,14 @@ output_can_use_rotation (output_t *output, Rotation rotation)
     return True;
 }
 
+static Bool
+output_is_primary(output_t *output)
+{
+    if (has_1_3)
+	    return XRRGetOutputPrimary(dpy, root) == output->output.xid;
+    return False;
+}
+
 static void
 set_output_info (output_t *output, RROutput xid, XRROutputInfo *output_info)
 {
@@ -1028,10 +1062,14 @@ set_output_info (output_t *output, RROutput xid, XRROutputInfo *output_info)
 	else
 	    init_transform (&output->transform);
     }
+
+    /* set primary */
+    if (!(output->changes & changes_primary))
+	output->primary = output_is_primary(output);
 }
     
 static void
-get_screen (void)
+get_screen (Bool current)
 {
     if (!has_1_2)
         fatal ("Server RandR version before 1.2\n");
@@ -1039,7 +1077,10 @@ get_screen (void)
     XRRGetScreenSizeRange (dpy, root, &minWidth, &minHeight,
 			   &maxWidth, &maxHeight);
     
-    res = XRRGetScreenResources (dpy, root);
+    if (current)
+	res = XRRGetScreenResourcesCurrent (dpy, root);
+    else
+	res = XRRGetScreenResources (dpy, root);
     if (!res) fatal ("could not get screen resources");
 }
 
@@ -1050,7 +1091,7 @@ get_crtcs (void)
 
     num_crtcs = res->ncrtc;
     crtcs = calloc (num_crtcs, sizeof (crtc_t));
-    if (!crtcs) fatal ("out of memory");
+    if (!crtcs) fatal ("out of memory\n");
     
     for (c = 0; c < res->ncrtc; c++)
     {
@@ -1058,15 +1099,22 @@ get_crtcs (void)
 #if RANDR_MAJOR > 1 || RANDR_MINOR >= 3
 	XRRCrtcTransformAttributes  *attr;
 #endif
-	int	    x;
 	XRRPanning  *panning_info = NULL;
 
-	if (has_1_3)
+	if (has_1_3) {
+	    XRRPanning zero;
+	    memset(&zero, 0, sizeof(zero));
 	    panning_info = XRRGetPanning  (dpy, res, res->crtcs[c]);
+	    zero.timestamp = panning_info->timestamp;
+	    if (!memcmp(panning_info, &zero, sizeof(zero))) {
+		Xfree(panning_info);
+		panning_info = NULL;
+	    }
+	}
 
 	set_name_xid (&crtcs[c].crtc, res->crtcs[c]);
 	set_name_index (&crtcs[c].crtc, c);
-	if (!crtc_info) fatal ("could not get crtc 0x%x information", res->crtcs[c]);
+	if (!crtc_info) fatal ("could not get crtc 0x%x information\n", res->crtcs[c]);
 	crtcs[c].crtc_info = crtc_info;
 	crtcs[c].panning_info = panning_info;
 	if (crtc_info->mode == None)
@@ -1108,7 +1156,7 @@ crtc_add_output (crtc_t *crtc, output_t *output)
 	crtc->mode_info = output->mode_info;
 	copy_transform (&crtc->pending_transform, &output->transform);
    }
-    if (!crtc->outputs) fatal ("out of memory");
+    if (!crtc->outputs) fatal ("out of memory\n");
     crtc->outputs[crtc->noutput++] = output;
 }
 
@@ -1142,6 +1190,82 @@ set_panning (void)
     }
 }
 
+static void
+set_gamma(void)
+{
+    output_t	*output;
+
+    for (output = outputs; output; output = output->next) {
+	int i, size;
+	crtc_t *crtc;
+	XRRCrtcGamma *gamma;
+
+	if (!(output->changes & changes_gamma))
+	    continue;
+
+	if (!output->crtc_info) {
+	    fatal("Need crtc to set gamma on.\n");
+	    continue;
+	}
+
+	crtc = output->crtc_info;
+
+	size = XRRGetCrtcGammaSize(dpy, crtc->crtc.xid);
+
+	if (!size) {
+	    fatal("Gamma size is 0.\n");
+	    continue;
+	}
+
+	gamma = XRRAllocGamma(size);
+	if (!gamma) {
+	    fatal("Gamma allocation failed.\n");
+	    continue;
+	}
+
+	for (i = 0; i < size; i++) {
+	    if (output->gamma.red == 1.0)
+		gamma->red[i] = i << 8;
+	    else
+		gamma->red[i] = (CARD16)(pow((double)i/(double)(size - 1),
+			    (double)output->gamma.red) * (double)(size - 1) * 256);
+
+	    if (output->gamma.green == 1.0)
+		gamma->green[i] = i << 8;
+	    else
+		gamma->green[i] = (CARD16)(pow((double)i/(double)(size - 1),
+			    (double)output->gamma.green) * (double)(size - 1) * 256);
+
+	    if (output->gamma.blue == 1.0)
+		gamma->blue[i] = i << 8;
+	    else
+		gamma->blue[i] = (CARD16)(pow((double)i/(double)(size - 1),
+			    (double)output->gamma.blue) * (double)(size - 1) * 256);
+	}
+
+	XRRSetCrtcGamma(dpy, crtc->crtc.xid, gamma);
+
+	free(gamma);
+    }
+}
+
+static void
+set_primary(void)
+{
+    output_t *output;
+
+    if (no_primary) {
+	XRRSetOutputPrimary(dpy, root, None);
+    } else {
+	for (output = outputs; output; output = output->next) {
+	    if (!(output->changes & changes_primary))
+		continue;
+	    if (output->primary)
+		XRRSetOutputPrimary(dpy, root, output->output.xid);
+	}
+    }
+}
+
 static Status
 crtc_disable (crtc_t *crtc)
 {
@@ -1226,7 +1350,7 @@ crtc_apply (crtc_t *crtc)
 	    if (has_1_3)
 		s = XRRSetPanning (dpy, res, crtc->crtc.xid, crtc->panning_info);
 	    else
-		fatal ("panning needs RandR 1.3");
+		fatal ("panning needs RandR 1.3\n");
 	}
     }
     free (rr_outputs);
@@ -1307,12 +1431,21 @@ panic (Status s, crtc_t *crtc)
     exit (1);
 }
 
-void
+static void
 apply (void)
 {
     Status  s;
     int	    c;
     
+    /*
+     * Hold the server grabbed while messing with
+     * the screen so that apps which notice the resize
+     * event and ask for xinerama information from the server
+     * receive up-to-date information
+     */
+    if (grab_server)
+	XGrabServer (dpy);
+    
     /*
      * Turn off any crtcs which are to be disabled or which are
      * larger than the target size
@@ -1359,14 +1492,6 @@ apply (void)
 	    panic (s, crtc);
     }
 
-    /*
-     * Hold the server grabbed while messing with
-     * the screen so that apps which notice the resize
-     * event and ask for xinerama information from the server
-     * receive up-to-date information
-     */
-    XGrabServer (dpy);
-    
     /*
      * Set the screen size
      */
@@ -1384,17 +1509,21 @@ apply (void)
 	if (s != RRSetConfigSuccess)
 	    panic (s, crtc);
     }
+
+    set_primary ();
+
     /*
      * Release the server grab and let all clients
      * respond to the updated state
      */
-    XUngrabServer (dpy);
+    if (grab_server)
+	XUngrabServer (dpy);
 }
 
 /*
  * Use current output state to complete the output list
  */
-void
+static void
 get_outputs (void)
 {
     int		o;
@@ -1404,7 +1533,7 @@ get_outputs (void)
 	XRROutputInfo	*output_info = XRRGetOutputInfo (dpy, res, res->outputs[o]);
 	output_t	*output;
 	name_t		output_name;
-	if (!output_info) fatal ("could not get output 0x%x information", res->outputs[o]);
+	if (!output_info) fatal ("could not get output 0x%x information\n", res->outputs[o]);
 	set_name_xid (&output_name, res->outputs[o]);
 	set_name_index (&output_name, o);
 	set_name_string (&output_name, output_info->name);
@@ -1468,7 +1597,7 @@ get_outputs (void)
     }
 }
 
-void
+static void
 mark_changing_crtcs (void)
 {
     int	c;
@@ -1501,8 +1630,8 @@ mark_changing_crtcs (void)
 /*
  * Test whether 'crtc' can be used for 'output'
  */
-Bool
-check_crtc_for_output (crtc_t *crtc, output_t *output, Bool ignore_state)
+static Bool
+check_crtc_for_output (crtc_t *crtc, output_t *output)
 {
     int		c;
     int		l;
@@ -1533,9 +1662,6 @@ check_crtc_for_output (crtc_t *crtc, output_t *output, Bool ignore_state)
 	    return False;
     }
 
-    if (ignore_state)
-	return True;
-
     if (crtc->noutput)
     {
 	/* make sure the state matches */
@@ -1567,7 +1693,7 @@ check_crtc_for_output (crtc_t *crtc, output_t *output, Bool ignore_state)
     return True;
 }
 
-crtc_t *
+static crtc_t *
 find_crtc_for_output (output_t *output)
 {
     int	    c;
@@ -1579,7 +1705,7 @@ find_crtc_for_output (output_t *output)
 	crtc = find_crtc_by_xid (output->output_info->crtcs[c]);
 	if (!crtc) fatal ("cannot find crtc 0x%x\n", output->output_info->crtcs[c]);
 
-	if (check_crtc_for_output (crtc, output, False))
+	if (check_crtc_for_output (crtc, output))
 	    return crtc;
     }
     return NULL;
@@ -1748,8 +1874,8 @@ set_screen_size (void)
 }
     
 #endif
-    
-void
+
+static void
 disable_outputs (output_t *outputs)
 {
     while (outputs)
@@ -1762,7 +1888,7 @@ disable_outputs (output_t *outputs)
 /*
  * find the best mapping from output to crtc available
  */
-int
+static int
 pick_crtcs_score (output_t *outputs)
 {
     output_t	*output;
@@ -1799,7 +1925,7 @@ pick_crtcs_score (output_t *outputs)
 	
 	/* reset crtc allocation for following outputs */
 	disable_outputs (outputs);
-	if (!check_crtc_for_output (crtc, output, True))
+	if (!check_crtc_for_output (crtc, output))
 	    continue;
 	
 	my_score = 1000;
@@ -1828,7 +1954,7 @@ pick_crtcs_score (output_t *outputs)
 /*
  * Pick crtcs for any changing outputs that don't have one
  */
-void
+static void
 pick_crtcs (void)
 {
     output_t	*output;
@@ -1910,6 +2036,7 @@ main (int argc, char **argv)
     Bool	propit = False;
     Bool	query_1 = False;
     int		major, minor;
+    Bool	current = False;
 #endif
 
     program_name = argv[0];
@@ -1933,6 +2060,16 @@ main (int argc, char **argv)
 	    verbose = True;
 	    continue;
 	}
+	if (!strcmp ("--nograb", argv[i])) {
+	    grab_server = False;
+	    continue;
+	}
+	if (!strcmp("--current", argv[i])) {
+	    current = True;
+	    /* if --current was the only arg, then query */
+	    if (argc == 2) query = True;
+	    continue;
+	}
 
 	if (!strcmp ("-s", argv[i]) || !strcmp ("--size", argv[i])) {
 	    if (++i>=argc) usage ();
@@ -2151,6 +2288,28 @@ main (int argc, char **argv)
 	    output->changes |= changes_panning;
 	    continue;
 	}
+	if (!strcmp ("--gamma", argv[i])) {
+	    if (!output) usage();
+	    if (++i>=argc) usage ();
+	    if (sscanf(argv[i], "%f:%f:%f", &output->gamma.red, 
+		    &output->gamma.green, &output->gamma.blue) != 3)
+		usage ();
+	    output->changes |= changes_gamma;
+	    setit_1_2 = True;
+	    continue;
+	}
+	if (!strcmp ("--primary", argv[i])) {
+	    if (!output) usage();
+	    output->changes |= changes_primary;
+	    output->primary = True;
+	    setit_1_2 = True;
+	    continue;
+	}
+	if (!strcmp ("--noprimary", argv[i])) {
+	    no_primary = True;
+	    setit_1_2 = True;
+	    continue;
+	}
 	if (!strcmp ("--set", argv[i])) {
 	    output_prop_t   *prop;
 	    if (!output) usage();
@@ -2364,6 +2523,8 @@ main (int argc, char **argv)
 	if (setit && !setit_1_2)
 	    query_1 = True;
     }
+    if (version)
+	printf("xrandr program version       " VERSION "\n");
 
     dpy = XOpenDisplay (display_name);
 
@@ -2396,7 +2557,7 @@ main (int argc, char **argv)
     {
 	umode_t	*m;
 
-        get_screen ();
+        get_screen (current);
 	get_crtcs();
 	get_outputs();
 	
@@ -2444,7 +2605,7 @@ main (int argc, char **argv)
     if (has_1_2 && propit)
     {
 	
-        get_screen ();
+        get_screen (current);
 	get_crtcs();
 	get_outputs();
 	
@@ -2507,6 +2668,8 @@ main (int argc, char **argv)
 		    nelements = strlen (prop->value);
 		    format = 8;
 		}
+		else
+		    continue;
 		XRRChangeOutputProperty (dpy, output->output.xid,
 					 name, type, format, PropModeReplace,
 					 data, nelements);
@@ -2520,7 +2683,7 @@ main (int argc, char **argv)
     }
     if (setit_1_2)
     {
-	get_screen ();
+	get_screen (current);
 	get_crtcs ();
 	get_outputs ();
 	set_positions ();
@@ -2594,7 +2757,12 @@ main (int argc, char **argv)
 	 * Set panning
 	 */
 	set_panning ();
-	
+
+	/* 
+	 * Set gamma on crtc's that belong to the outputs.
+	 */
+	set_gamma ();
+
 	/*
 	 * Now apply all of the changes
 	 */
@@ -2610,7 +2778,7 @@ main (int argc, char **argv)
 	
 #define ModeShown   0x80000000
 	
-	get_screen ();
+	get_screen (current);
 	get_crtcs ();
 	get_outputs ();
 
@@ -2642,7 +2810,7 @@ main (int argc, char **argv)
 			    mode->width, mode->height, output->x, output->y);
 		}
 		if (verbose)
-		    printf (" (0x%x)", mode->id);
+		    printf (" (0x%x)", (int)mode->id);
 		if (output->rotation != RR_Rotate_0 || verbose)
 		{
 		    printf (" %s", 
@@ -2678,7 +2846,7 @@ main (int argc, char **argv)
 	    if (mode)
 	    {
 		printf (" %dmm x %dmm",
-			output_info->mm_width, output_info->mm_height);
+			(int)output_info->mm_width, (int)output_info->mm_height);
 	    }
 
 	    if (crtc && crtc->panning_info && crtc->panning_info->width > 0)
@@ -2706,8 +2874,8 @@ main (int argc, char **argv)
 
 	    if (verbose)
 	    {
-		printf ("\tIdentifier: 0x%x\n", output->output.xid);
-		printf ("\tTimestamp:  %d\n", output_info->timestamp);
+		printf ("\tIdentifier: 0x%x\n", (int)output->output.xid);
+		printf ("\tTimestamp:  %d\n", (int)output_info->timestamp);
 		printf ("\tSubpixel:   %s\n", order[output_info->subpixel_order]);
 		printf ("\tClones:    ");
 		for (j = 0; j < output_info->nclone; j++)
@@ -2789,26 +2957,35 @@ main (int argc, char **argv)
 		    } else if (actual_type == XA_INTEGER &&
 			       actual_format == 32)
 		    {
-			printf("\t%s: %d (0x%08x)",
-			       XGetAtomName (dpy, props[j]),
-			       *(INT32 *)prop, *(INT32 *)prop);
+			printf("\t%s: ", XGetAtomName (dpy, props[j]));
+			for (k = 0; k < nitems; k++) {
+			    if (k > 0)
+				printf ("\n\t\t\t");
+			    printf("%d (0x%08x)",
+				   (int)((INT32 *)prop)[k], (int)((INT32 *)prop)[k]);
+			}
 
  			if (propinfo->range && propinfo->num_values > 0) {
-			    printf(" range%s: ",
+			    if (nitems > 1)
+				printf ("\n\t\t");
+			    printf("\trange%s: ",
 				   (propinfo->num_values == 2) ? "" : "s");
 
 			    for (k = 0; k < propinfo->num_values / 2; k++)
-				printf(" (%d,%d)", propinfo->values[k * 2],
-				       propinfo->values[k * 2 + 1]);
+				printf(" (%d,%d)", (int)propinfo->values[k * 2],
+				       (int)propinfo->values[k * 2 + 1]);
 			}
 
 			printf("\n");
 		    } else if (actual_type == XA_ATOM &&
 			       actual_format == 32)
 		    {
-			printf("\t%s: %s",
-			       XGetAtomName (dpy, props[j]),
-			       XGetAtomName (dpy, *(Atom *)prop));
+			printf("\t%s:", XGetAtomName (dpy, props[j]));
+			for (k = 0; k < nitems; k++) {
+			    if (k > 0 && (k & 1) == 0)
+				printf ("\n\t\t");
+			    printf("\t%s", XGetAtomName (dpy, ((Atom *)prop)[k]));
+			}
 
  			if (!propinfo->range && propinfo->num_values > 0) {
 			    printf("\n\t\tsupported:");
@@ -2823,13 +3000,13 @@ main (int argc, char **argv)
 			}
 			printf("\n");
 		    } else if (actual_format == 8) {
-			printf ("\t\t%s: %s%s\n", XGetAtomName (dpy, props[j]),
+			printf ("\t%s: %s%s\n", XGetAtomName (dpy, props[j]),
 				prop, bytes_after ? "..." : "");
 		    } else {
 			char	*type = actual_type ? XGetAtomName (dpy, actual_type) : "none";
-			printf ("\t\t%s: %s(%d) (format %d items %d) ????\n",
+			printf ("\t%s: %s(%d) (format %d items %d) ????\n",
 				XGetAtomName (dpy, props[j]),
-				type, actual_type, actual_format, nitems);
+				type, (int)actual_type, actual_format, (int)nitems);
 		    }
 
 		    free(propinfo);
@@ -2844,7 +3021,7 @@ main (int argc, char **argv)
 		    int		f;
 		    
 		    printf ("  %s (0x%x) %6.1fMHz",
-			    mode->name, mode->id,
+			    mode->name, (int)mode->id,
 			    (float)mode->dotClock / 1000000.0);
 		    for (f = 0; mode_flags[f].flag; f++)
 			if (mode->modeFlags & mode_flags[f].flag)
@@ -2905,7 +3082,7 @@ main (int argc, char **argv)
 	    if (!(mode->modeFlags & ModeShown))
 	    {
 		printf ("  %s (0x%x) %6.1fMHz\n",
-			mode->name, mode->id,
+			mode->name, (int)mode->id,
 			(float)mode->dotClock / 1000000.0);
 		printf ("        h: width  %4d start %4d end %4d total %4d skew %4d clock %6.1fKHz\n",
 			mode->width, mode->hSyncStart, mode->hSyncEnd,
@@ -3052,7 +3229,7 @@ main (int argc, char **argv)
     if (setit && !dryrun) XRRSelectInput (dpy, root,
 			       RRScreenChangeNotifyMask);
     if (setit && !dryrun) status = XRRSetScreenConfigAndRate (dpy, sc,
-						   DefaultRootWindow (dpy), 
+						   root,
 						   (SizeID) size, (Rotation) (rotation | reflection), rate, CurrentTime);
 
     XRRQueryExtension(dpy, &event_base, &error_base);
